/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <tiuser.h>
#include <signal.h>

typedef struct creq {
	struct creq *next;
	struct creq *prev;
	struct t_call *callp;
	int afd;
} creq_t;

creq_t *cfree;
creq_t *chead;
creq_t *ctail;
struct t_discon *disp;

void dolook(int fd);

extern int t_errno;
extern char *t_errlist[];

extern void log(const char *, ...);

#define tlog(STR) log("%s: %s", (STR), t_errlist[t_errno])

void
serve(int fd, int qlen, char *transport)
{
	creq_t *creqp;
	struct t_call *callp;
	pid_t pid;
	int i;

	/*
	 * Allocate a t_discon structure to be used when
	 * retrieving disconnect indications from the stream.
	 */
	disp = (struct t_discon *)t_alloc(fd, T_DIS, T_ALL);
	if (disp == NULL) {
		tlog("t_alloc failed");
		exit(1);
	}

	/*
	 * Allocate an array of t_call structures.  (qlen
	 * specifies the number of structures to allocate.)
	 * For each structure, initialize it and then link
	 * it onto the freelist.
	 */
	creqp = calloc(qlen, sizeof(creq_t));
	if (creqp == NULL) {
		log("calloc failed");
		exit(1);
	}
	for (i = 0; i < qlen; i++, creqp++) {
		callp = (struct t_call *)t_alloc(fd, T_CALL,
		  T_ALL);
		if (callp == NULL) {
			tlog("t_alloc failed");
			exit(1);
		}
		creqp->callp = callp;
		creqp->afd = -1;
		creqp->next = cfree;
		cfree = creqp;
	}
	creqp = NULL;
	for ( ; ; ) {
		/*
		 * Use the current connect request structure
		 * in processing the next connect indication.
		 * If we have no current structure, pull one
		 * off of the freelist.
		 */
		if (creqp == NULL) {
			creqp = cfree;
			cfree = cfree->next;
		}
		callp = creqp->callp;

		/*
		 * Wait for a connect indication to arrive.
		 */
		if (t_listen(fd, callp) < 0) {
			tlog("t_listen failed");
			exit(1);
		}

		/*
		 * Get a new transport endpoint to be used in
		 * the connection.
		 */
		if ((creqp->afd = getfd(transport)) < 0) {
			/*
			 * Can't get a new endpoint -- reject
			 * the connect request.
			 */
			if (t_snddis(fd, callp) < 0) {
				tlog("t_snddis failed");
				exit(1);
			}
			continue;
		}
		while (creqp != NULL) {
			/*
			 * Accept the connection on creqp->afd.
			 */
			if (t_accept(fd, creqp->afd, callp) < 0) {
				if (t_errno != TLOOK) {
					tlog("t_accept failed");
					exit(1);
				}

				/*
				 * A pending event is preventing t_accept
				 * from succeeding.  Save creqp on the
				 * list of pending connections and call
				 * dolook to process the event.
				 */
				creqp->next = chead;
				creqp->prev = NULL;
				if (chead != NULL)
					chead->prev = creqp;
				else
					ctail = creqp;
				chead = creqp;
				dolook(fd);
			} else {
				/*
				 * Accept succeeded. Fork and let the
				 * child process do the real work.  The
				 * parent goes back to listen for more
				 * connect indications.
				 */
				if ((pid = fork()) < 0) {
					/*
					 * Fork failed -- reject
					 * the connect request.
					 */
					log("fork failed");
					if (t_snddis(creqp->afd, NULL) < 0) {
						tlog("t_snddis failed");
						exit(1);
					}
				} else if (pid == 0) {	/* child */
					/*
					 * Child resumes here.  Don't need
					 * the file descriptor used to listen
					 * for connect indications.
					 */
					t_close(fd);

					/*
					 * Perform the service using
					 * creqp->afd and exit.
					 */
				}

				/*
				 * After a successful fork, the parent
				 * continues here. Close the file
				 * descriptor used in the connection,
				 * place the creq structure back on the
				 * freelist, and continue to listen for
				 * more connect indications.
				 */
				t_close(creqp->afd);
				creqp->afd = -1;
				creqp->next = cfree;
				cfree = creqp;
			}

			/*
			 * Process the next creq structure, if there
			 * is one, on the pending list.
			 */
			creqp = chead;
			if (creqp != NULL) {
				callp = creqp->callp;
				chead = creqp->next;
				if (chead != NULL)
					chead->prev = NULL;
				else
					ctail = NULL;
			}
		}
	}
}

void
dolook(int fd)
{
	creq_t *creqp;
	struct t_call *callp;
	int event;
	char buf[64];

	switch (event = t_look(fd)) {
	case T_LISTEN:
		/*
		 * Another connect indication has arrived.
		 * Take a creq structure off the freelist and
		 * remove the connect indication from the stream.
		 */
		creqp = cfree;
		cfree = cfree->next;
		if (t_listen(fd, creqp->callp) < 0) {
			tlog("t_listen failed");
			exit(1);
		}

		/*
		 * Put the connect indication on the end of
		 * the pending list of connections.
		 */
		creqp->next = NULL;
		creqp->prev = ctail;
		if (ctail != NULL)
			ctail->next = creqp;
		else
			chead = creqp;
		ctail = creqp;
		break;

	case T_DISCONNECT:
		/*
		 * Someone canceled a connect request.  Remove
		 * the disconnect indication from the stream and
		 * find a pending connection that matches it.
		 */
		if (t_rcvdis(fd, disp) < 0) {
			tlog("t_rcvdis failed");
			exit(1);
		}
		for (creqp = chead; creqp; creqp = creqp->next) {
			callp = creqp->callp;
			if (callp->sequence == disp->sequence) {
				/*
				 * We found a match.  Close the transport
				 * endpoint allocated for the connection,
				 * unlink the creq structure from the list
				 * of pending connections, and place it on
				 * the freelist.
				 */
				t_close(creqp->afd);
				if (creqp->prev != NULL)
					creqp->prev->next = creqp->next;
				else
					chead = creqp->next;
				if (creqp->next != NULL)
					creqp->next->prev = creqp->prev;
				else
					ctail = creqp->prev;
				creqp->next = cfree;
				cfree = creqp;
				break;
			}
		}
		break;

	case T_ERROR:
	case -1:
		tlog("t_look failed");
		exit(1);

	default:
		/*
		 * This shouldn't happen.
		 */
		sprintf(buf, "unknown t_look event %d", event);
		log(buf);
		exit(1);
	}
}
